﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Xml.Linq;
using DocumentFormat.OpenXml.Packaging;
using OpenXmlPowerTools;

class Program
{
    static void Main(string[] args)
    {
        var source1 = "../../Source1.docx";
        var source2 = "../../Source2.docx";
        var noNumbering1 = "../../Source1-no-numbering.docx";
        var noNumbering2 = "../../Source2-no-numbering.docx";
        var merged = "../../Merged.docx";
        var mergedAndNumbered = "../../MergedAndNumbered.docx";

        var fis1 = new FileInfo(source1);
        var fis2 = new FileInfo(source2);
        var finn1 = new FileInfo(noNumbering1);
        var finn2 = new FileInfo(noNumbering2);

        File.Copy(fis1.FullName, finn1.FullName, true);
        File.Copy(fis2.FullName, finn2.FullName, true);

        // Remove numbering from the heading styles (if numbering exists)
        RemoveHeadingNumbering(finn1.FullName);
        RemoveHeadingNumbering(finn2.FullName);

        // Use DocumentBuilder and merge the files
        var sources = new List<Source>() {
            new Source(new WmlDocument(finn1.FullName), false),
            new Source(new WmlDocument(finn2.FullName), false),
        };
        var mergedDocument = DocumentBuilder.BuildDocument(sources);
        mergedDocument.SaveAs(merged);

        // Add numbering to heading styles
        var m1 = new FileInfo(merged);
        var m2 = new FileInfo(mergedAndNumbered);
        File.Copy(m1.FullName, m2.FullName, true);
        AddHeadingNumbering(m2.FullName);
    }

    private static void RemoveHeadingNumbering(string fileName)
    {
        using (WordprocessingDocument wDoc = WordprocessingDocument.Open(fileName, true))
        {
            var styleXDocument = wDoc.MainDocumentPart.StyleDefinitionsPart.GetXDocument();
            var stylesToAlter = styleXDocument
                .Root
                .Elements(W.style)
                .Where(s =>
                {
                    if ((string)s.Attribute(W.type) != "paragraph")
                        return false;
                    var styleId = (string)s.Attribute(W.styleId);
                    if (styleId.StartsWith("Heading") &&
                        styleId.Length == 8 &&
                        char.IsDigit(styleId[7]))
                        return true;
                    return false;
                })
                .ToList();
            foreach (var style in stylesToAlter)
            {
                style.Elements(W.pPr).Elements(W.numPr).Remove();
            }
            wDoc.MainDocumentPart.StyleDefinitionsPart.PutXDocument();
        }
    }

    private static void AddHeadingNumbering(string fileName)
    {
        using (WordprocessingDocument wDoc = WordprocessingDocument.Open(fileName, true))
        {
            // get the next abstractNumId
            // if there is no numbering part, then the next abstract Id = 0
            int abstractNumId = 0;
            var numberingPart = wDoc
                .MainDocumentPart
                .NumberingDefinitionsPart;
            if (numberingPart != null)
            {
                abstractNumId = numberingPart
                    .GetXDocument()
                    .Root
                    .Elements(W.abstractNum)
                    .Select(a => (int)a.Attribute(W.abstractNumId))
                    .Max() + 1;
            }

            // get the next numId
            // if there is no numbering part, then the next numId = 1;
            int numId = 1;
            if (numberingPart != null)
            {
                numId = numberingPart
                    .GetXDocument()
                    .Root
                    .Elements(W.num)
                    .Select(n => (int)n.Attribute(W.numId))
                    .Max() + 1;
            }

            // if there is no numbering part, then add it
            XDocument numberingXDoc = null;
            if (numberingPart == null)
            {
                numberingPart = wDoc.MainDocumentPart.AddNewPart<NumberingDefinitionsPart>();
                numberingXDoc = new XDocument(
                    new XElement(W.numbering,
                        new XAttribute(XNamespace.Xmlns + "w", W.w)));
            }
            else
            {
                numberingXDoc = numberingPart.GetXDocument();
            }

            // add the new numbering to the numbering definitions part
            // create the new abstract numbering definition
            var abstractNumDef = XElement.Parse(
                string.Format(@"<w:abstractNum w:abstractNumId='{0}' xmlns:w='http://schemas.openxmlformats.org/wordprocessingml/2006/main'>
  <w:nsid w:val='41F645DF'/>
  <w:multiLevelType w:val='multilevel'/>
  <w:tmpl w:val='04090025'/>
  <w:lvl w:ilvl='0'>
    <w:start w:val='1'/>
    <w:numFmt w:val='upperRoman'/>
    <w:pStyle w:val='Heading1'/>
    <w:lvlText w:val='Chapter %1'/>
    <w:lvlJc w:val='left'/>
    <w:pPr>
      <w:ind w:left='432'
              w:hanging='432'/>
    </w:pPr>
  </w:lvl>
  <w:lvl w:ilvl='1'>
    <w:start w:val='1'/>
    <w:numFmt w:val='decimal'/>
    <w:pStyle w:val='Heading2'/>
    <w:lvlText w:val='%1.%2'/>
    <w:lvlJc w:val='left'/>
    <w:pPr>
      <w:ind w:left='576'
              w:hanging='576'/>
    </w:pPr>
  </w:lvl>
  <w:lvl w:ilvl='2'>
    <w:start w:val='1'/>
    <w:numFmt w:val='decimal'/>
    <w:pStyle w:val='Heading3'/>
    <w:lvlText w:val='%1.%2.%3'/>
    <w:lvlJc w:val='left'/>
    <w:pPr>
      <w:ind w:left='720'
              w:hanging='720'/>
    </w:pPr>
  </w:lvl>
  <w:lvl w:ilvl='3'>
    <w:start w:val='1'/>
    <w:numFmt w:val='decimal'/>
    <w:pStyle w:val='Heading4'/>
    <w:lvlText w:val='%1.%2.%3.%4'/>
    <w:lvlJc w:val='left'/>
    <w:pPr>
      <w:ind w:left='864'
              w:hanging='864'/>
    </w:pPr>
  </w:lvl>
  <w:lvl w:ilvl='4'>
    <w:start w:val='1'/>
    <w:numFmt w:val='decimal'/>
    <w:pStyle w:val='Heading5'/>
    <w:lvlText w:val='%1.%2.%3.%4.%5'/>
    <w:lvlJc w:val='left'/>
    <w:pPr>
      <w:ind w:left='1008'
              w:hanging='1008'/>
    </w:pPr>
  </w:lvl>
  <w:lvl w:ilvl='5'>
    <w:start w:val='1'/>
    <w:numFmt w:val='decimal'/>
    <w:pStyle w:val='Heading6'/>
    <w:lvlText w:val='%1.%2.%3.%4.%5.%6'/>
    <w:lvlJc w:val='left'/>
    <w:pPr>
      <w:ind w:left='1152'
              w:hanging='1152'/>
    </w:pPr>
  </w:lvl>
  <w:lvl w:ilvl='6'>
    <w:start w:val='1'/>
    <w:numFmt w:val='decimal'/>
    <w:pStyle w:val='Heading7'/>
    <w:lvlText w:val='%1.%2.%3.%4.%5.%6.%7'/>
    <w:lvlJc w:val='left'/>
    <w:pPr>
      <w:ind w:left='1296'
              w:hanging='1296'/>
    </w:pPr>
  </w:lvl>
  <w:lvl w:ilvl='7'>
    <w:start w:val='1'/>
    <w:numFmt w:val='decimal'/>
    <w:pStyle w:val='Heading8'/>
    <w:lvlText w:val='%1.%2.%3.%4.%5.%6.%7.%8'/>
    <w:lvlJc w:val='left'/>
    <w:pPr>
      <w:ind w:left='1440'
              w:hanging='1440'/>
    </w:pPr>
  </w:lvl>
  <w:lvl w:ilvl='8'>
    <w:start w:val='1'/>
    <w:numFmt w:val='decimal'/>
    <w:pStyle w:val='Heading9'/>
    <w:lvlText w:val='%1.%2.%3.%4.%5.%6.%7.%8.%9'/>
    <w:lvlJc w:val='left'/>
    <w:pPr>
      <w:ind w:left='1584'
              w:hanging='1584'/>
    </w:pPr>
  </w:lvl>
</w:abstractNum>", abstractNumId));
            abstractNumDef.Attributes(XNamespace.Xmlns + "w").Remove();

            var numDef = XElement.Parse(
                string.Format(@"<w:num w:numId='{1}' xmlns:w='http://schemas.openxmlformats.org/wordprocessingml/2006/main'>
  <w:abstractNumId w:val='{0}'/>
</w:num>", abstractNumId, numId));
            numDef.Attributes(XNamespace.Xmlns + "w").Remove();

            var abstractToAddAfter = numberingXDoc
                .Root
                .Elements(W.abstractNum)
                .LastOrDefault();
            if (abstractToAddAfter != null)
                abstractToAddAfter.AddAfterSelf(abstractNumDef);
            else
                numberingXDoc.Root.Add(abstractNumDef);

            var numToAddAfter = numberingXDoc
                .Root
                .Elements(W.num)
                .LastOrDefault();
            if (numToAddAfter != null)
                numToAddAfter.AddAfterSelf(numDef);
            else
                numberingXDoc.Root.Add(numDef);

            numberingPart.PutXDocument(numberingXDoc);

            // add appropriate numpr to styles
            var styleXDocument = wDoc.MainDocumentPart.StyleDefinitionsPart.GetXDocument();
            var stylesToAlter = styleXDocument
                .Root
                .Elements(W.style)
                .Where(s =>
                {
                    if ((string)s.Attribute(W.type) != "paragraph")
                        return false;
                    var styleId = (string)s.Attribute(W.styleId);
                    if (styleId.StartsWith("Heading") &&
                        styleId.Length == 8 &&
                        char.IsDigit(styleId[7]))
                        return true;
                    return false;
                })
                .ToList();
            foreach (var style in stylesToAlter)
            {
                var id = (string)style.Attribute(W.styleId);
                int lvl;
                if (Int32.TryParse(id.Substring(7), out lvl))
                {
                    lvl -= 1;
                    var newNumPr = new XElement(W.numPr,
                        new XElement(W.ilvl,
                            new XAttribute(W.val, lvl)),
                        new XElement(W.numId,
                            new XAttribute(W.val, numId)));
                    var pPr = style.Element(W.pPr);
                    if (pPr != null)
                        pPr.Add(newNumPr);
                }
            }
            wDoc.MainDocumentPart.StyleDefinitionsPart.PutXDocument();
        }
    }

}
